package be.iminds.ilabt.jfed.experimenter_gui.editor.views;

import be.iminds.ilabt.jfed.experimenter_gui.editor.ToolBoxFilter;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ToolBoxFilterBuilder;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.ObservableList;
import javafx.geometry.Side;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.layout.*;
import javafx.scene.paint.Color;
import org.controlsfx.glyphfont.FontAwesome;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

public class ToolBoxFilterPane extends VBox {
    private final Label filtersLabel;
    private final MenuButton editFiltersButton;
    private final FlowPane filterFlowPanel;

    private final EditToolBoxFiltersDialogFactory editToolBoxFiltersDialogFactory;
    private ObjectProperty<ToolBoxFilter> filter =
            new SimpleObjectProperty<>(this, "filter", new ToolBoxFilter());

    private final Map<FilterPart, Node> filterPartToNodeMap = new HashMap<>();

    public ToolBoxFilterPane(EditToolBoxFiltersDialogFactory editToolBoxFiltersDialogFactory) {
        this.editToolBoxFiltersDialogFactory = editToolBoxFiltersDialogFactory;
        getStylesheets().add(getClass().getResource("ExperimenterEditor.css").toExternalForm());
        getStyleClass().add("toolbox-content");

        HBox header = new HBox();

        this.filtersLabel = new Label("");
        filtersLabel.getStyleClass().add("toolbox-filter-label");

        Pane expandingPane = new Pane();
        HBox.setHgrow(expandingPane, Priority.ALWAYS);

        MenuItem addTestbedFilterMI = new MenuItem("Add Testbed Filter...");
        MenuItem addTestbedFromMapFilterMI = new MenuItem("Add Testbed Filter using map...",
                GlyphUtils.createGlyph(FontAwesome.Glyph.MAP_MARKER));
        MenuItem addFederationFilterMI = new MenuItem("Add Federation Filter...");
        MenuItem addOrganisationFilterMI = new MenuItem("Add Organisation Filter...");
        MenuItem addCategoryFilterMI = new MenuItem("Add Category Filter...");
        MenuItem clearFilterMI = new MenuItem("Clear all Filters");

        this.editFiltersButton = new MenuButton("+", null,
                addTestbedFilterMI, addTestbedFromMapFilterMI, addFederationFilterMI, addOrganisationFilterMI, addCategoryFilterMI,
                new SeparatorMenuItem(),
                clearFilterMI);
        editFiltersButton.getStyleClass().remove("menu-button");
        editFiltersButton.getStyleClass().add("button");
        editFiltersButton.setPopupSide(Side.RIGHT);
        editFiltersButton.getStyleClass().add("toolbox-filter-edit");

        addTestbedFilterMI.setOnAction(
                (e) -> {
                    setFilter(editToolBoxFiltersDialogFactory.showEditTestbedsDialog(this.getScene().getWindow(), filter.get()));
                }
        );
        addTestbedFromMapFilterMI.setOnAction(
                (e) -> {
                    setFilter(editToolBoxFiltersDialogFactory.showAddTestbedFromMapDialog(this.getScene().getWindow(), filter.get()));
                }
        );
        addFederationFilterMI.setOnAction(
                (e) -> {
                    setFilter(editToolBoxFiltersDialogFactory.showEditFederationDialog(this.getScene().getWindow(), filter.get()));
                }
        );
        addOrganisationFilterMI.setOnAction(
                (e) -> {
                    setFilter(editToolBoxFiltersDialogFactory.showEditOrganisationDialog(this.getScene().getWindow(), filter.get()));
                }
        );
        addCategoryFilterMI.setOnAction(
                (e) -> {
                    setFilter(editToolBoxFiltersDialogFactory.showEditCategoryDialog(this.getScene().getWindow(), filter.get()));
                }
        );
        clearFilterMI.setOnAction(
                (e) -> {
                    setFilter(new ToolBoxFilter());
                }
        );

        header.getChildren().addAll(filtersLabel, expandingPane, editFiltersButton);

        filterFlowPanel = new FlowPane();
        filterFlowPanel.setHgap(3.0);
        filterFlowPanel.setVgap(3.0);
        filterFlowPanel.setPrefWrapLength(200.0);
        filterFlowPanel.setPrefWidth(210.0);
        filterFlowPanel.setMaxWidth(210.0);

        //flowpanel

//        Button filter1 = new Button("vwall2");
//        filter1.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.TIMES, Color.DARKGRAY));
//        filter1.getStyleClass().add("toolbox-filter-filter");
//
//        Button filter2 = new Button("wilab2");
//        filter2.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.TIMES, Color.DARKGRAY));
//        filter2.getStyleClass().add("toolbox-filter-filter");
//
//        Button filter3 = new Button("org: iminds");
//        filter3.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.TIMES, Color.DARKGRAY));
//        filter3.getStyleClass().add("toolbox-filter-filter");
//
//        Button filter4 = new Button("way too long filter description that is way too long");
//        filter4.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.TIMES, Color.DARKGRAY));
//        filter4.getStyleClass().add("toolbox-filter-filter");
//
//
//        filters.getChildren().addAll(
//                new FlowPanelBugWorkaroundPane(180.0, filter1),
//                new FlowPanelBugWorkaroundPane(180.0, filter2),
//                new FlowPanelBugWorkaroundPane(180.0, filter3),
//                new FlowPanelBugWorkaroundPane(180.0, filter4)
//        );

        setSpacing(5.0);
        getChildren().addAll(header, filterFlowPanel);

        filter.addListener((observable, oldValue, newValue) -> updateFilters(newValue));
        updateFilters(filter.get());
    }

    public final void setFilter(ToolBoxFilter value) { filterProperty().set(value); }
    public final ToolBoxFilter getFilter() {return filter.get(); }
    public ObjectProperty<ToolBoxFilter> filterProperty() { return filter; }

    public void clearFilters() {
        filterFlowPanel.getChildren().removeAll(filterPartToNodeMap.values());
    }

    public void updateFilters(@Nullable ToolBoxFilter filter) {
        clearFilters();
        if (filter != null) {
            if (filter.isEmpty())
                filtersLabel.setText("No Filters");
            else
                filtersLabel.setText("Active Filters:");

            filterFlowPanel.getChildren().addAll(
                    filterToParts(filter)
                            .stream()
                            .map(this::createFilterPartButton)
                            .collect(Collectors.toList())
            );
        }
    }

    private Node createFilterPartButton(@Nonnull FilterPart filterPart) {
        String prefix;
        switch (filterPart.type) {
            case TESTBED: { prefix = ""; break; }
            case ORGANISATION: { prefix = "org: "; break; }
            case FEDERATION: { prefix = "fed: "; break; }
            case CATEGORY: { prefix = "cat: "; break; }
            default: throw new IllegalArgumentException("filterPart.type="+filterPart.type+" is not supported");
        }
        Button res = new Button(prefix + filterPart.getVal());
        res.setGraphic(GlyphUtils.createGlyph(FontAwesome.Glyph.TIMES, Color.DARKGRAY));
        res.getStyleClass().add("toolbox-filter-filter");
        res.setOnAction((e) -> this.removeFilterPart(filterPart));

        Node rres = new FlowPanelBugWorkaroundPane(180.0, res);
        filterPartToNodeMap.put(filterPart, rres);
        return rres;
    }

    private void removeFilterPart(@Nonnull FilterPart filterPart) {
        assert filter.get() != null;
        ToolBoxFilterBuilder newFilter = new ToolBoxFilterBuilder(filter.get());
        switch (filterPart.type) {
            case TESTBED: { newFilter.removeTestbedId(filterPart.val); break; }
            case ORGANISATION: { newFilter.removeOrganisationId(filterPart.val);  break; }
            case FEDERATION: { newFilter.removeFederationId(filterPart.val);  break; }
            case CATEGORY: { newFilter.removeCategoryName(filterPart.val);  break; }
            default: throw new IllegalArgumentException("filterPart.type="+filterPart.type+" is not supported");
        }
        setFilter(newFilter.create());
    }

    private static List<FilterPart> filterToParts(ToolBoxFilter filter) {
        List<FilterPart> res = new ArrayList<>();
        filter.getTestbedIds().stream()
                .map(name -> new FilterPart(FilterPartType.TESTBED, name))
                .forEach(res::add);
        filter.getCategoryNames().stream()
                .map(name -> new FilterPart(FilterPartType.CATEGORY, name))
                .forEach(res::add);
        filter.getOrganisationIds().stream()
                .map(name -> new FilterPart(FilterPartType.ORGANISATION, name))
                .forEach(res::add);
        filter.getFederationIds().stream()
                .map(name -> new FilterPart(FilterPartType.FEDERATION, name))
                .forEach(res::add);
        return res;
    }

    private enum FilterPartType {
        TESTBED, ORGANISATION, FEDERATION, CATEGORY
    }
    private static class FilterPart {
        private final FilterPartType type;
        private final String val;

        public FilterPart(FilterPartType type, String val) {
            this.type = type;
            this.val = val;
        }

        public FilterPartType getType() {
            return type;
        }

        public String getVal() {
            return val;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;
            FilterPart that = (FilterPart) o;
            return type == that.type &&
                    Objects.equals(val, that.val);
        }

        @Override
        public int hashCode() {
            return Objects.hash(type, val);
        }
    }


    /**
     * There's an anoying but with FlowPane, where the the width of the FlowPane
     * exceeds its maxWidth, if one of it's children has a prefWidth that is larger than the FlowPane's max width.
     * Setting the child's maxWidth does not help.
     *
     * This workaround seems to work...
     */
    public static class FlowPanelBugWorkaroundPane extends Pane {
        public FlowPanelBugWorkaroundPane() {
            super();
        }

        public FlowPanelBugWorkaroundPane(Node... children) {
            super(children);
        }
        public FlowPanelBugWorkaroundPane(double maxWidth, Node... children) {
            super(children);
            setMaxWidth(maxWidth);
            for (Node child : children) {
                if (child instanceof Region) {
                    ((Region)child).setMaxWidth(maxWidth);
                }
            }
        }

        @Override
        protected double computePrefWidth(double height) {
            double orig = super.computePrefWidth(height);
            if (orig > getMaxWidth())
                return getMaxWidth();
            return orig;
        }
    }
}
